__version__ = "20.52.0"

# TODO: Maybe figure out how to include the 'pyproject.toml' file as part of the package installation files.
# import toml

# PY_PROJECT_FILE = './pyproject.toml'

# parsed_tom = toml.load(PY_PROJECT_FILE)

# __version__ = parsed_tom['tool']['poetry']['version']

# print(__version__)

